package cz.nic.tablexia;

import com.badlogic.gdx.Application;
import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.graphics.GL30;

import net.engio.mbassy.listener.Handler;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.Locale;

import cz.nic.tablexia.bus.ApplicationBus;
import cz.nic.tablexia.bus.ApplicationBus.ApplicationEvent;
import cz.nic.tablexia.bus.event.LocaleChangedEvent;
import cz.nic.tablexia.loader.application.ApplicationFontManager;
import cz.nic.tablexia.loader.application.ApplicationSoundManager;
import cz.nic.tablexia.loader.application.ApplicationTextManager;
import cz.nic.tablexia.loader.application.ApplicationTextureManager;
import cz.nic.tablexia.loader.zip.ZipAssetLoader;
import cz.nic.tablexia.menu.MainMenuContainer;
import cz.nic.tablexia.screen.AbstractTablexiaScreen;
import cz.nic.tablexia.screen.loader.LoaderScreen;
import cz.nic.tablexia.util.Log;
import cz.nic.tablexia.util.Utility;

public class Tablexia extends TablexiaApplication {

    private boolean loadingComplete = false;
    private MainMenuContainer mainMenuContainer;
    private ZipAssetLoader    zipAssetLoader;
    private boolean           resetState;

    public Tablexia(boolean debug, Locale systemLocale, String versionName, boolean resetState) {
        TablexiaSettings.init(debug, systemLocale, versionName);
        this.resetState = resetState;
    }

    public Tablexia(String buildTypeKey, Locale systemLocale, String versionName, boolean resetState) {
        TablexiaSettings.init(buildTypeKey, systemLocale, versionName);
        this.resetState = resetState;
    }

    private void loadingComplete() {
        if (!loadingComplete) {
            loadingComplete = true;
            showLastOrInitialScreen();
            ApplicationBus.getInstance().publishAsync(new ApplicationLoadingCompleteEvent());
        }
    }

    private void showLastOrInitialScreen() {
        AbstractTablexiaScreen<?> lastScreen = TablexiaSettings.getInstance().getCurrentScreen();
        if (lastScreen != null) {
            setScreenIfIsDifferent(lastScreen, ScreenTransaction.FADE);
        } else {
            setScreenIfIsDifferent(Utility.getScreenForScreenClass(TablexiaSettings.INITIAL_SCREEN), ScreenTransaction.MOVE_LEFT);
        }
    }

    private void prepareMainMenu() {
        mainMenuContainer = new MainMenuContainer(getStage().getWidth(), getStage().getHeight());
        getStage().addActor(mainMenuContainer);
    }

    private void startLoading(Locale locale) {
        // sync loaded screen with loader image
        setScreenIfIsDifferent(new LoaderScreen());

        // async internal assets loading
        ApplicationFontManager.getInstance().load();
        ApplicationTextManager.getInstance().load(locale);
        // async zip extraction
        zipAssetLoader = new ZipAssetLoader();
        zipAssetLoader.load(locale);
        // async external assets loading
        ApplicationTextureManager.getInstance().load();
        ApplicationSoundManager.getInstance().load();

        loadingComplete = false;
    }

    private void clearLoadedData() {
        // clear main menu
        if (mainMenuContainer != null) {
            mainMenuContainer.remove();
            mainMenuContainer = null;
        }

        // dispose all loaders
        ApplicationFontManager.getInstance().dispose();
        ApplicationTextManager.getInstance().dispose();
        zipAssetLoader.dispose();
        zipAssetLoader = null;
        ApplicationTextureManager.getInstance().dispose();
        ApplicationSoundManager.getInstance().dispose();
    }

//////////////////////////// LOCALE CHANGED HANDLER

    @Handler
    public void handleLocaleChangedEvent(LocaleChangedEvent localeChangedEvent) {
        Gdx.app.postRunnable(new Runnable() {

            @Override
            public void run() {
                clearLoadedData();
                startLoading(TablexiaSettings.getInstance().getLocale());
            }
        });
    }


//////////////////////////// LIBGDX LIFECYCLE

    @Override
    public void create() {
        super.create();

        TablexiaSettings.getInstance().loadPreferences(resetState);
        Log.setLoglevel(TablexiaSettings.getInstance().getLogLevel());

        // init event bus handlers
        ApplicationBus.getInstance().subscribe(this);

        // start loading application scope data
        startLoading(TablexiaSettings.getInstance().getLocale());
    }

    @Override
    public void pause() {
        if (mainMenuContainer != null && getScreen() != null && getScreen().canBePaused()) {
            // don't use menu open animation for android and iOS pause event (animation is run after resume)
            mainMenuContainer.openMainMenu(Gdx.app.getType().equals(Application.ApplicationType.Desktop));
        }
        super.pause();
    }


    @Override
    public void render() {
        // render other screens
        super.render();

        if (screenshotListener != null) {
            obtainScreenshot();
            screenshotListener = null;
        }

        // process loading
        if (!loadingComplete) {
            // load internal assets
            if (!ApplicationFontManager.getInstance().update()) return;
            if (!ApplicationTextManager.getInstance().update()) return;
            // load external assets
            if (!zipAssetLoader.update()) return;
            if (!ApplicationTextureManager.getInstance().update()) return;
            if (!ApplicationSoundManager.getInstance().update()) return;

            prepareMainMenu();
            loadingComplete();
        }
    }

    ScreenshotListener screenshotListener;

    /**
     * Queues request for screenshot with this listener, it is called after obtaining screenshot.
     *
     * @param listener
     */
    public void takeScreenShot(ScreenshotListener listener) {
        this.screenshotListener = listener;
    }

    /**
     * Listener interface, that gets called after creating screenshot array
     */
    public static interface ScreenshotListener {
        public void screenshotTaken(int[] screen);
    }

    /**
     * Method launched during render phase before cleaning the buffer, creates bytearray of screen pixels and passes it to the listener
     */
    private void obtainScreenshot() {
        Gdx.gl.glFinish();
        int width = Gdx.graphics.getWidth();
        int height = Gdx.graphics.getHeight();
        int bt[] = new int[width * height];

        ByteBuffer pbuf = ByteBuffer.allocateDirect(width * height * 4);
        pbuf.order(ByteOrder.nativeOrder());
        Gdx.gl.glReadPixels(0, 0, width, height, GL30.GL_RGBA, GL30.GL_UNSIGNED_BYTE, pbuf);
        pbuf.asIntBuffer().get(bt);
        if (screenshotListener != null) {
            screenshotListener.screenshotTaken(bt);
        }
    }

    @Override
    public void dispose() {
        super.dispose();
        clearLoadedData();
    }


//////////////////////////// APPLICATION LOADING COMPLETE EVENT

    public static class ApplicationLoadingCompleteEvent implements ApplicationEvent {
    }


//////////////////////////// CHANGE SCREEN EVENT

    public static class ChangeScreenEvent implements ApplicationEvent {

        private Class<? extends AbstractTablexiaScreen<?>> screen;
        private ScreenTransaction                          screenTransaction;

        public ChangeScreenEvent(Class<? extends AbstractTablexiaScreen<?>> screen, ScreenTransaction screenTransaction) {
            this.screen = screen;
            this.screenTransaction = screenTransaction;
        }

        public Class<? extends AbstractTablexiaScreen<?>> getScreen() {
            return screen;
        }

        public ScreenTransaction getScreenTransaction() {
            return screenTransaction;
        }
    }

    @Handler
    public void handleChangeScreenEvent(final ChangeScreenEvent changeScreenEvent) {
        final Class<? extends AbstractTablexiaScreen<?>> screenClass = changeScreenEvent.getScreen();
        if (!loadingComplete) {
            Log.err(((Object) this).getClass(), "Cannot change screen -> Application loading not complete!");
            return;
        }
        if (screenClass == null) {
            Log.err(((Object) this).getClass(), "Cannot change screen -> Received empty screen class!");
            return;
        }
        // create new screen on GL thread
        Gdx.app.postRunnable(new Runnable() {

            @Override
            public void run() {
                setScreenIfIsDifferent(Utility.getScreenForScreenClass(screenClass), changeScreenEvent.getScreenTransaction());
            }
        });
    }
}
