package cz.nic.tablexia;

import com.badlogic.gdx.ApplicationListener;
import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.InputMultiplexer;
import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.graphics.GL20;
import com.badlogic.gdx.graphics.Texture;
import com.badlogic.gdx.graphics.g2d.Batch;
import com.badlogic.gdx.math.Interpolation;
import com.badlogic.gdx.scenes.scene2d.Actor;
import com.badlogic.gdx.scenes.scene2d.Stage;
import com.badlogic.gdx.utils.viewport.ScreenViewport;

import cz.nic.tablexia.bus.ApplicationBus;
import cz.nic.tablexia.bus.ApplicationBus.ApplicationEvent;
import cz.nic.tablexia.debug.DebugInfo;
import cz.nic.tablexia.graphics.ColorTexture;
import cz.nic.tablexia.screen.AbstractTablexiaScreen;
import cz.nic.tablexia.screen.AbstractTablexiaScreen.ScreenLoadingListener;

import static com.badlogic.gdx.scenes.scene2d.actions.Actions.alpha;
import static com.badlogic.gdx.scenes.scene2d.actions.Actions.fadeIn;
import static com.badlogic.gdx.scenes.scene2d.actions.Actions.fadeOut;
import static com.badlogic.gdx.scenes.scene2d.actions.Actions.moveTo;
import static com.badlogic.gdx.scenes.scene2d.actions.Actions.run;
import static com.badlogic.gdx.scenes.scene2d.actions.Actions.sequence;

/**
 * Abstract Tablexia libGDX application screenshotListener with screen transactions implementation.
 *
 * @author Matyáš Latner
 */
public abstract class TablexiaApplication implements ApplicationListener {

    private static final Color BACKGROUND_COLOR = new Color(0f, 0f, 0f, 1f);

    private AbstractTablexiaScreen<?> lastScreen;
    private AbstractTablexiaScreen<?> screen;
    private InputMultiplexer inputMultiplexer;
    private DebugInfo debugInfo;
    private Stage stage;


    public Stage getStage() {
        return stage;
    }


//////////////////////////// LIBGDX LIFECYCLE

    @Override
    public void dispose() {
        inputMultiplexer.removeProcessor(stage);
        if (TablexiaSettings.getInstance().isDebug()) {
            debugInfo.dispose();
        }
        stage.dispose();
        if (lastScreen != null) {
            lastScreen.hide();
            lastScreen.dispose();
        }
        if (screen != null) {
            screen.hide();
            screen.dispose();
        }
    }

    @Override
    public void pause() {
        if (lastScreen != null) lastScreen.pause();
        if (screen != null) screen.pause();
    }

    @Override
    public void resume() {
        if (lastScreen != null) lastScreen.resume();
        if (screen != null) screen.resume();
    }

    @Override
    public void create() {
        stage = new Stage(new ScreenViewport());
        inputMultiplexer = new InputMultiplexer(stage);
        Gdx.input.setInputProcessor(inputMultiplexer);
        prepareDebugInfo();
    }

    @Override
    public void render() {
        // clear screen every frame
        Gdx.gl.glClearColor(BACKGROUND_COLOR.r, BACKGROUND_COLOR.g, BACKGROUND_COLOR.b, BACKGROUND_COLOR.a);
        Gdx.gl.glClear(GL20.GL_COLOR_BUFFER_BIT);

        if (TablexiaSettings.getInstance().isDebug()) {
            debugInfo.update();
        }
        float deltaTime = Gdx.graphics.getDeltaTime();
        if (lastScreen != null) lastScreen.render(deltaTime);
        if (screen != null) screen.render(deltaTime);

        stage.act(deltaTime);
        stage.draw();
    }

    @Override
    public void resize(int width, int height) {
        if (lastScreen != null) lastScreen.resize(width, height);
        if (screen != null) screen.resize(width, height);
        stage.getViewport().update(width, height, true);
        // keep debug info on edges of the screen
        if (TablexiaSettings.getInstance().isDebug()) {
            debugInfo.onResize();
        }
    }


//////////////////////////// DEBUG INFO

    private void prepareDebugInfo() {
        if (TablexiaSettings.getInstance().isDebug()) {
            debugInfo = new DebugInfo(); // TODO probably deprecated creating specific size
            getStage().addActor(debugInfo);
        }
    }


//////////////////////////// SCREEN HANDLING

    public static class ScreenChangedEvent implements ApplicationEvent {

        private Class<? extends AbstractTablexiaScreen<?>> screenClass;

        public ScreenChangedEvent(Class<? extends AbstractTablexiaScreen<?>> screenClass) {
            this.screenClass = screenClass;
        }

        public Class<? extends AbstractTablexiaScreen<?>> getScreenClass() {
            return screenClass;
        }
    }

    /**
     * Change current screen if is different from actual screen.
     *
     * @param screen new screen to change for
     */
    public void setScreenIfIsDifferent(AbstractTablexiaScreen<?> screen) {
        if (getScreen() == null || getScreen().getClass() != screen.getClass()) {
            setScreen(screen);
        }
    }

    /**
     * Change current screen.
     *
     * @param screen new screen to change for
     */
    public void setScreen(AbstractTablexiaScreen<?> screen) {
        lastScreen = this.screen;
        this.screen = screen;
        processLastScreen();
        processNewScreen(screen);
    }

    /**
     * Change current screen to new screen with screen transaction.
     * If new screen is same as current screen no change is performed.
     * If {@link ScreenTransaction} is <code>null</code> screen is changed without transaction.
     *
     * @param newScreen         new screen to change for
     * @param screenTransaction screen transaction type
     */
    public void setScreenIfIsDifferent(AbstractTablexiaScreen<?> newScreen, ScreenTransaction screenTransaction) {
        if (newScreen != null) {
            if (getScreen() == null || getScreen().getClass() != newScreen.getClass()) {
                setScreen(newScreen, screenTransaction);
            }
        }
    }

    /**
     * Change current screen to new screen with screen transaction.
     * If {@link ScreenTransaction} is <code>null</code> screen is changed without transaction.
     *
     * @param newScreen         new screen to change for
     * @param screenTransaction screen transaction type
     */
    public void setScreen(final AbstractTablexiaScreen<?> newScreen, ScreenTransaction screenTransaction) {
        if (screenTransaction == null) {
            setScreen(newScreen);
        } else {
            lastScreen = screen;
            screen = newScreen;
            screenTransaction.processTransaction(stage, lastScreen, newScreen,

                    new Runnable() {

                        @Override
                        public void run() {
                            processNewScreen(screen);
                        }
                    },

                    new Runnable() {
                        @Override
                        public void run() {
                            processLastScreen();
                        }
                    },

                    new Runnable() {
                        @Override
                        public void run() {
                            newScreen.performScreenVisible();
                        }
                    });
        }
    }

    /**
     * Return current screen
     *
     * @return current screen
     */
    public AbstractTablexiaScreen<?> getScreen() {
        return screen;
    }

    @SuppressWarnings("unchecked")
    private void processNewScreen(AbstractTablexiaScreen<?> newScreen) {
        if (newScreen != null) {
            TablexiaSettings.getInstance().setCurrentScreen(newScreen);
            inputMultiplexer.addProcessor(newScreen.getInputProcessor());
            newScreen.show();
            newScreen.resize(Gdx.graphics.getWidth(), Gdx.graphics.getHeight());
            ApplicationBus.getInstance().post(new ScreenChangedEvent((Class<? extends AbstractTablexiaScreen<?>>) newScreen.getClass())).asynchronously();
        }
    }

    private void processLastScreen() {
        if (lastScreen != null) {
            inputMultiplexer.removeProcessor(lastScreen.getInputProcessor());
            lastScreen.hide();
            lastScreen.dispose();
            lastScreen = null;
            System.gc();
        }
    }


//////////////////////////// SCREEN TRANSACTIONS

    public enum ScreenTransaction {

        MOVE_RIGHT(new MoveRightAnimation()),
        MOVE_LEFT(new MoveLeftAnimation()),
        FADE(new FadeAnimation());

        private interface ScreenTransactionImplementation {
            public void processTransaction(Stage stage, AbstractTablexiaScreen<?> lastScreen, AbstractTablexiaScreen<?> newScreen, Runnable newScreenHandler, Runnable lastScreenHandler, Runnable finishHandler);
        }

        private ScreenTransactionImplementation screenTransactionImplementation;

        private ScreenTransaction(ScreenTransactionImplementation screenTransactionImplementation) {
            this.screenTransactionImplementation = screenTransactionImplementation;
        }

        private void processTransaction(Stage stage, AbstractTablexiaScreen<?> lastScreen, AbstractTablexiaScreen<?> newScreen, Runnable newScreenHandler, Runnable lastScreenHandler, Runnable finishHandler) {
            screenTransactionImplementation.processTransaction(stage, lastScreen, newScreen, newScreenHandler, lastScreenHandler, finishHandler);
        }
    }

    private static class MoveRightAnimation implements cz.nic.tablexia.TablexiaApplication.ScreenTransaction.ScreenTransactionImplementation {

        protected float newScreenMoveFrom = -Gdx.graphics.getWidth();
        protected float newScreenMoveTo = 0;
        protected float lastScreenMoveTo = Gdx.graphics.getWidth();

        @Override
        public void processTransaction(Stage stage, final AbstractTablexiaScreen<?> lastScreen, final AbstractTablexiaScreen<?> newScreen, Runnable newScreenHandler, final Runnable lastScreenHandler, final Runnable finishHandler) {
            final int transactionSpeed = 2;
            newScreen.setLoadingListener(new ScreenLoadingListener() {

                @Override
                public void loadingComplete() {
                    newScreen.resetLoadingListener();
                    if (lastScreen != null) {
                        lastScreen.addAction(sequence(moveTo(lastScreenMoveTo, 0, transactionSpeed, Interpolation.pow4Out), run(lastScreenHandler)));
                    }
                    if (newScreen != null) {
                        newScreen.setPosition(newScreenMoveFrom, 0);
                        newScreen.addAction(sequence(moveTo(newScreenMoveTo, 0, transactionSpeed, Interpolation.pow4Out), run(finishHandler)));
                    }
                }
            });
            newScreenHandler.run();
        }
    }

    private static class MoveLeftAnimation extends MoveRightAnimation {

        @Override
        public void processTransaction(Stage stage, final AbstractTablexiaScreen<?> lastScreen, AbstractTablexiaScreen<?> newScreen, Runnable newScreenHandler, final Runnable lastScreenHandler, final Runnable finishHandler) {
            newScreenMoveFrom = Gdx.graphics.getWidth();
            newScreenMoveTo = 0;
            lastScreenMoveTo = -Gdx.graphics.getWidth();
            super.processTransaction(stage, lastScreen, newScreen, newScreenHandler, lastScreenHandler, finishHandler);
        }
    }

    private static class FadeAnimation implements cz.nic.tablexia.TablexiaApplication.ScreenTransaction.ScreenTransactionImplementation {

        @Override
        public void processTransaction(Stage stage, final AbstractTablexiaScreen<?> lastScreen, final AbstractTablexiaScreen<?> newScreen, final Runnable newScreenHandler, final Runnable lastScreenHandler, final Runnable finishHandler) {
            final int transactionSpeed = 1;
            final Texture texture = new ColorTexture(Gdx.graphics.getWidth(), Gdx.graphics.getHeight(), Color.BLACK);
            final Actor screenDimmer = new Actor() {

                public void draw(Batch batch, float parentAlpha) {
                    batch.setColor(getColor());
                    batch.draw(texture, 0, 0);
                    batch.setColor(Color.WHITE);
                }

                ;
            };
            newScreen.setLoadingListener(new ScreenLoadingListener() {

                @Override
                public void loadingComplete() {
                    newScreen.resetLoadingListener();
                    screenDimmer.addAction(sequence(fadeOut(transactionSpeed, Interpolation.pow4In), run(finishHandler), run(new Runnable() {
                        public void run() {
                            screenDimmer.remove();
                        }
                    })));
                }
            });
            screenDimmer.addAction(alpha(0));
            screenDimmer.addAction(sequence(fadeIn(transactionSpeed, Interpolation.pow4Out), run(newScreenHandler), run(lastScreenHandler)));
            stage.addActor(screenDimmer);
        }
    }

}
