package cz.nic.tablexia.screen;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.InputProcessor;
import com.badlogic.gdx.ScreenAdapter;
import com.badlogic.gdx.audio.Sound;
import com.badlogic.gdx.files.FileHandle;
import com.badlogic.gdx.graphics.Texture;
import com.badlogic.gdx.graphics.g2d.BitmapFont;
import com.badlogic.gdx.scenes.scene2d.Action;
import com.badlogic.gdx.scenes.scene2d.Stage;
import com.badlogic.gdx.utils.I18NBundle;
import com.badlogic.gdx.utils.async.AsyncTask;

import net.engio.mbassy.listener.Handler;

import java.util.ArrayList;
import java.util.List;
import java.util.Locale;

import cz.nic.tablexia.TablexiaSettings;
import cz.nic.tablexia.bus.ApplicationBus;
import cz.nic.tablexia.bus.ApplicationBus.ApplicationEvent;
import cz.nic.tablexia.bus.event.MainMenuPositionChangedEvent;
import cz.nic.tablexia.loader.TablexiaAssetManager;
import cz.nic.tablexia.loader.TablexiaDataManager;
import cz.nic.tablexia.loader.TablexiaSoundManager;
import cz.nic.tablexia.loader.TablexiaTextureManager;
import cz.nic.tablexia.loader.application.ApplicationFontManager;
import cz.nic.tablexia.util.Log;
import cz.nic.tablexia.util.ui.XFillViewport;

import static com.badlogic.gdx.scenes.scene2d.actions.Actions.moveTo;

public abstract class AbstractTablexiaScreen<T> extends ScreenAdapter {

    private Stage 					stage;
    private TablexiaTextureManager  textureManager;
    private TablexiaSoundManager    soundManager;
    private TextManager 			textManager;
    private DataManager 			dataManager;
    private ScreenLoadingListener 	screenLoadingListener;
    private boolean                 loadingComplete;
    private boolean                 loadingStarted;
    private boolean                 loadAsync;
    private boolean                 paused;
    private boolean                 hasState;

    private static class TextManager extends TablexiaDataManager<I18NBundle> {

		private static class TextLoader implements AsyncTask<I18NBundle> {

			private String textResourceFileName;
			private Locale locale;

			public TextLoader(String textResourceFileName, Locale locale) {
				this.textResourceFileName = textResourceFileName;
				this.locale = locale;
			}

			@Override
			public I18NBundle call() throws Exception {
				FileHandle baseFileHandle = Gdx.files.internal(textResourceFileName);
				return I18NBundle.createBundle(baseFileHandle, locale);
			}
		}
		
		public void load(Locale locale, String textFileName) {
			if (textFileName != null) {				
				setAsyncTask(new TextLoader(textFileName, locale));
			}
		}
	}
	
	private class DataManager extends TablexiaDataManager<T> {

		public class DataLoader implements AsyncTask<T> {

			@Override
			public T call() throws Exception {
				return prepareScreenData();
			}
		}
		
		public void load() {
			setAsyncTask(new DataLoader());
		}
	}
	
	public static interface ScreenLoadingListener {
		void loadingComplete();
	}
	
	public static class ScreenInfoEvent implements ApplicationEvent {

		private String infoKey;
		private String infoValue;

		public ScreenInfoEvent(String infoKey, String infoValue) {
			this.infoKey = infoKey;
			this.infoValue = infoValue;
		}
		
		public String getInfoKey() {
			return infoKey;
		}
		
		public String getInfoValue() {
			return infoValue;
		}
	}


    public AbstractTablexiaScreen() {
        this(true, true, TablexiaAssetManager.StorageType.EXTERNAL);
    }
	
	public AbstractTablexiaScreen(boolean hasState, boolean loadAsync, TablexiaAssetManager.StorageType storageType) {
        stage               = prepareStage();
        this.hasState       = hasState;
        this.loadAsync      = loadAsync;
		textureManager 	    = new TablexiaTextureManager(storageType);
		soundManager 	    = new TablexiaSoundManager(storageType);
		textManager 	    = new TextManager();
		dataManager 	    = new DataManager();

		loadingComplete = false;
        loadingStarted = false;
	}

    public boolean hasState() {
        return hasState;
    }

//////////////////////////// SCREEN PAUSE

    public boolean canBePaused() {
        return false;
    }

//////////////////////////// LOADING LISTENER

    private void notifyLoadingCompleteListeners() {
        if (screenLoadingListener != null) {
            screenLoadingListener.loadingComplete();
        }
    }

    public void setLoadingListener(ScreenLoadingListener listener) {
        screenLoadingListener = listener;
    }

    public void resetLoadingListener() {
        screenLoadingListener = null;
    }


//////////////////////////// STAGE

    private Stage prepareStage() {
		return new Stage(new XFillViewport());
    }

    public Stage getStage() {
        return stage;
    }

    public void setPosition(float positionX, float positionY) {
        stage.addAction(moveTo(positionX, positionY));
    }

    public void addAction(Action action) {
        stage.addAction(action);
    }


//////////////////////////// INPUT PROCESSORS

    public InputProcessor getInputProcessor() {
        return getStage();
    }


//////////////////////////// MAIN MENU POSITION CHANGE HANDLER

    @Handler
    public void handleMainMenuPositionChangedEvent(MainMenuPositionChangedEvent mainMenuPositionChangedEvent) {
        paused = mainMenuPositionChangedEvent.isMenuOpen();
    }


//////////////////////////// LIBGDX LIFECYCLE
	
	@Override
	public final void show() {
        ApplicationBus.getInstance().subscribe(this);
		if (!loadingComplete) {
			startLoading();
		}
	}

    @Override
    public final void hide() {
        ApplicationBus.getInstance().unsubscribe(this);
    }

    @Override
	public final void resize(int width, int height) {
		stage.getViewport().update(width, height, true);
		if (loadingComplete) {
			performScreenResized(width, height);
		}
	}

	@Override
	public final void render(float delta) {
        if (!loadingComplete) {
            // SCREEN LOADING
            if (loadingStarted) {

                if (loadAsync) {
                    if (!textureManager.update()) return;
                    if (!soundManager.update()) return;
                    if (!textManager.update()) return;
                    if (!dataManager.update()) return;
                } else {
                    textureManager.finishLoading();
                    soundManager.finishLoading();
                    textManager.finishLoading();
                    dataManager.finishLoading();
                }

                loadingComplete = true;
                loadingStarted = false;
                performScreenLoaded();
                notifyLoadingCompleteListeners();
            }
		} else {
            // SCREEN RENDERING
            screenRender(delta);
            if (!paused) {
                stage.act(delta);
            }
			stage.draw();
        }
	}
	
	@Override
	public final void dispose() {
        stage.dispose();
        textureManager.dispose();
        soundManager.dispose();
        textManager.dispose();
        dataManager.dispose();
        performScreenDisposed();
    }
	
	
//////////////////////////// TABLEXIA SCREEN LIFECYCLE
	
	private final void performScreenLoaded() {
		Log.info(getClass(), "[ ------- Screen Loaded ------- ]");
		screenLoaded();
	}
	
	public final void performScreenVisible() {
		Log.info(getClass(), "[ ------- Screen Visible ------- ]");
		screenVisible();
	}
	
	private final void performScreenResized(int width, int height) {
		Log.info(getClass(), "[ ------- Screen Resized ------- ]");
		screenResized(width, height);
	}
	
	private final void performScreenDisposed() {
		Log.info(getClass(), "[ ------- Screen Disposed ------- ]");
		screenDisposed();
	}
	
	protected void screenRender(float delta) {};
	protected void screenLoaded() {};
	protected void screenVisible() {};
	protected void screenDisposed() {};
	protected void screenResized(int width, int height) {};
	
	protected void 			prepareScreenTextureAssetNames(List<String> textureFileNames) {}
	protected void 			prepareScreenSoundAssetNames(List<String> soundsFileNames) {}
	protected String 		prepareScreenTextResourcesAssetName() {
		return null;
	}
	protected T 			prepareScreenData() {
		return null;
	}
	
	protected void 			printScreenInfo(String infoKey, String infoValue) {
		ApplicationBus.getInstance().post(new ScreenInfoEvent(infoKey, infoValue)).asynchronously();
	}


//////////////////////////// LOADERS

    protected void startLoading() {
        loadingComplete = false;
        loadingStarted  = true;

        startTextureLoader();
        startSoundLoader();
        startTextLoader();
        startDataLoader();
    }
	
	private void startTextureLoader() {
        List<String> textureFileNames = new ArrayList<String>();
        prepareScreenTextureAssetNames(textureFileNames);
        for (String textureFileName : textureFileNames) {
            textureManager.loadTexture(textureFileName);
        }
	}
	
	private void startSoundLoader() {
		List<String> soundsFileNames = new ArrayList<String>();
		prepareScreenSoundAssetNames(soundsFileNames);
		for (String soundFileName : soundsFileNames) {
			soundManager.loadSound(soundFileName);
		}
	}
	
	private void startTextLoader() {
		textManager.load(TablexiaSettings.getInstance().getLocale(), prepareScreenTextResourcesAssetName());
	}

    protected void startDataLoader() {
        dataManager.load();
    }

	
	public Texture getTexture(String textureName) {
		return textureManager.getTexture(textureName);
	}
	
	public Sound getSound(String soundName) {
		return soundManager.getSound(soundName);
	}
	
	public String getText(String key) {
		return textManager.getResult().get(key);
	}

    public BitmapFont getDefaultRegularFont() {
        return ApplicationFontManager.getInstance().getDefaultApplicationRegularFont();
    }

    public BitmapFont getDefaultBoldFont() {
        return ApplicationFontManager.getInstance().getDefaultApplicationBoldFont();
    }
	
	public String getFormattedText(String key, Object... args) {
		return textManager.getResult().format(key, args);
	}
	
	protected T getData() {
		return dataManager.getResult();
	}
}
